#!/usr/bin/env python3

import os
import ast
from datetime import datetime
from abc import ABC, abstractmethod

from tinydb import TinyDB

from das.modules.common import Logger


class AddOutputBase(ABC):
	"""Base class for updating DB with parsed portscan output."""

	def __init__(self, db, rm, scanner_name, scanner_args):
		"""
		Constructor.

		:param db: a tinydb database file path
		:type db: tinydb.TinyDB
		:param rm: a flag showing if we need to drop the DB before updating its values
		:type rm: bool
		:param scanner_name: name of the port scanner to run
		:type scanner_name: str
		:param scanner_args: port scanner arguments
		:type scanner_args: str
		:return: base class object
		:rtype: das.modules.add.AddOutputBase
		"""
		self.db = TinyDB(db)
		if rm:
			self.db.truncate()

		self.portscan_out = f'.db/raw/{scanner_name}-{datetime.now().strftime("%Y%m%dT%H%M%S")}.out'
		self.command = f"""sudo {scanner_name} {scanner_args} | tee {self.portscan_out}"""

		Logger.print_cmd(self.command)
		os.system(self.command)

		with open(self.portscan_out, 'r', encoding='utf-8') as fd:
			self.portscan_raw = fd.read().splitlines()

	@abstractmethod
	def parse(self):
		"""Interface for a parsing method."""
		raise NotImplementedError


class AddMasscanOutput(AddOutputBase):
	"""Child class for processing Masscan output."""

	def parse(self):
		"""
		Masscan raw output parser.

		:return: a pair of values (portscan raw output filename, number of hosts added to DB)
		:rtype: tuple
		"""
		hosts = set()
		for line in self.portscan_raw:
			try:
				ip = line.split()[-1]
				port, proto = line.split()[3].split('/')
			except:
				pass
			else:
				if proto == 'tcp':
					item = {'ip': ip, 'port': int(port)}
					if item not in self.db:
						self.db.insert(item)

				hosts.add(ip)

		return (self.portscan_out, len(hosts))


class AddRustscanOutput(AddOutputBase):
	"""Child class for processing RustScan output."""

	def parse(self):
		"""
		RustScan raw output parser.

		:return: a pair of values (portscan raw output filename, number of hosts added to DB)
		:rtype: tuple
		"""
		hosts = set()
		for line in self.portscan_raw:
			try:
				ip, ports = line.split(' -> ')
			except:
				pass
			else:
				for port in ast.literal_eval(ports):
					item = {'ip': ip, 'port': port}
					if item not in self.db:
						self.db.insert(item)

				hosts.add(ip)

		return (self.portscan_out, len(hosts))


class AddNaabuOutput(AddOutputBase):
	"""Child class for processing Naabu output."""

	def parse(self):
		"""
		Naabu raw output parser.

		:return: a pair of values (portscan raw output filename, number of hosts added to DB)
		:rtype: tuple
		"""
		hosts = set()
		for line in self.portscan_raw:
			try:
				ip, port = line.split(':')
			except:
				pass
			else:
				item = {'ip': ip, 'port': int(port)}
				if item not in self.db:
					self.db.insert(item)

				hosts.add(ip)

		return (self.portscan_out, len(hosts))


class AddNmapOutput(AddOutputBase):
	"""Child class for processing Nmap output."""

	def parse(self):
		"""
		Nmap raw output parser.

		:return: a pair of values (portscan raw output filename, number of hosts added to DB)
		:rtype: tuple
		"""
		hosts = set()
		for line in self.portscan_raw:
			try:
				ip = line.split()[-1]
				port, proto = line.split()[3].split('/')
			except:
				pass
			else:
				item = {'ip': ip, 'port': int(port)}
				if item not in self.db:
					self.db.insert(item)

				hosts.add(ip)

		return (self.portscan_out, len(hosts))



#!/usr/bin/env python3

# Based on: https://github.com/CBHue/nMap_Merger
# Original author: Hue B. Solutions LLC, CBHue

import os
import socket
from pathlib import Path
from collections import namedtuple

import defusedxml.ElementTree as ET
from netaddr import IPNetwork

from das.modules.common import Logger


class NmapMerger:
	"""Class for merging separate Nmap outputs into a single report in different formats."""

	def __init__(self, hosts, ports, output=None):
		"""
		Constructor.

		:param hosts: list of Nmap reports with host-like filenames to merge ("all" for all the reports in Nmap directory with host-like filenames)
		:type hosts: list
		:param ports: list of Nmap reports with port-like filenames to merge ("all" for all the reports in Nmap directory with port-like filenames)
		:type ports: list
		:param output: dictionary containing output type information and desired output filename
		:type output: dict
		:return: class object
		:rtype: das.modules.report.NmapMerger
		"""
		if output:
			for fmt, name in output.items():
				if name:
					Report = namedtuple('Report', 'format filename')
					self.output = Report(fmt, name)
		else:
			self.output = None

		P = (Path.cwd() / '.nmap').glob('*.*')
		P = list(P)

		if hosts:
			try:
				if Path(hosts).exists():
					with open(hosts, 'r') as fd:
						hosts = ','.join(i.strip() for i in fd.read().splitlines())
			except OSError:  # catching [Errno 36] File name too long
				pass

			if hosts == 'all':
				self.nmap_reports = {x for x in P if not x.stem.startswith('port')}
			else:
				hosts = hosts.split(',')
				hosts = [IPNetwork(h) for h in hosts]
				hosts = [str(ip).replace('.', '-') for ip_obj in hosts for ip in ip_obj]
				self.nmap_reports = {x for h in hosts for x in P if h == x.stem}

		elif ports:
			try:
				if Path(ports).exists():
					with open(ports, 'r') as fd:
						ports = ','.join(i.strip() for i in fd.read().splitlines())
			except OSError:  # catching [Errno 36] File name too long
				pass

			if ports == 'all':
				self.nmap_reports = {x for x in P if x.stem.startswith('port')}
			else:
				ports = ports.split(',')
				self.nmap_reports = {x for p in ports for x in P if p == x.stem}

	def xml_to_html(self, merged_xml):
		"""
		Convert merged Nmap XML report to HTML report.

		:param merged_xml: merged Nmap XML report filename
		:type merged_xml: str
		"""
		xsltproc = Path('/usr/bin/xsltproc')
		if xsltproc.exists():
			merged_html = f'{self.output.filename}.html'
			command = f'{str(xsltproc)} {merged_xml} -o {merged_html}'
			os.system(command)
			P = Path.cwd() / merged_html
			if P.exists():
				Logger.print_success(f'Merged HTML report -> {P.resolve()}')
		else:
			Logger.print_error(f'{xsltproc}: binary does not exist, install with "sudo apt install xsltproc -y"')

	def show(self):
		"""Print raw Nmap reports in simple text format."""
		text_reports = [r for r in self.nmap_reports if r.suffix == '.nmap']
		total_reports = len(text_reports)
		Logger.print_info(f'Total reports -> {total_reports}')

		try:
			sorted_reports = [str(r) for r in sorted(text_reports, key=lambda x: socket.inet_aton(x.stem.replace('-', '.')))]
		except OSError:
			sorted_reports = [str(r) for r in sorted(text_reports)]

		i = 1
		for report in sorted_reports:
			Logger.print_separator(report, prefix=f'{i}/{total_reports}')
			os.system(f'cat {report}')
			i += 1

	def generate(self):
		"""Perform all the steps needed to generate a single Nmap report."""
		if self.output.format in ('oX', 'oA'):
			merged_xml = f'{self.output.filename}.xml'
			for report in self.nmap_reports:
				if report.suffix == '.xml':
					d = self.get_header_dict(report)
					self.add_header(merged_xml, d)
					break

			Logger.print_info(f'Merging {len([r for r in self.nmap_reports if r.suffix == ".xml"])} XML reports')

			host_num = 0
			for report in self.nmap_reports:
				if report.suffix == '.xml':
					Logger.print_info(f'Processing XML report -> {report}')
					n = self.merge_nmap(report, merged_xml)
					host_num = host_num + n

			P = Path.cwd() / merged_xml
			if P.exists():
				Logger.print_success(f'Merged XML report -> {P.resolve()}')

			self.add_footer(merged_xml, host_num)
			self.xml_to_html(merged_xml)

		if self.output.format in ('oN', 'oA'):
			text_reports = [r for r in self.nmap_reports if r.suffix == '.nmap']

			try:
				sorted_reports = ' '.join(str(r) for r in sorted(text_reports, key=lambda x: socket.inet_aton(x.stem.replace('-', '.'))))
			except OSError:
				sorted_reports = ' '.join(str(r) for r in sorted(text_reports))

			if sorted_reports:
				merged_text = f'{self.output.filename}.nmap'
				os.system(f'cat {sorted_reports} > {merged_text}')
				P = Path.cwd() / merged_text
				if P.exists():
					Logger.print_success(f'Merged simple text report -> {P.resolve()}')
			else:
				Logger.print_error('No reports that meet the search criteria')

		if self.output.format in ('oG', 'oA'):
			grepable_reports = [r for r in self.nmap_reports if r.suffix == '.gnmap']

			try:
				sorted_reports = ' '.join(str(r) for r in sorted(grepable_reports, key=lambda x: socket.inet_aton(x.stem.replace('-', '.'))))
			except OSError:
				sorted_reports = ' '.join(str(r) for r in sorted(grepable_reports))

			if sorted_reports:
				merged_grepable = f'{self.output.filename}.gnmap'
				os.system(f'cat {sorted_reports} > {merged_grepable}')
				P = Path.cwd() / merged_grepable
				if P.exists():
					Logger.print_success(f'Merged grepable report -> {P.resolve()}')
			else:
				Logger.print_error('No reports that meet the search criteria')

	@staticmethod
	def get_header_dict(xml_report):
		"""
		Extract metadata from XML header of a raw Nmap XML report.

		:param xml_report: raw Nmap XML report filename
		:type xml_report: pathlib.PosixPath
		:return: a dictionary containing XML metadata
		:rtype: dict
		"""
		tree = ET.parse(xml_report)
		root = tree.getroot()
		nmaprun = root.attrib
		scaninfo = root.find('scaninfo').attrib
		merged = {**nmaprun, **scaninfo}

		d = {}
		for key, value in merged.items():
			d[key] = value.replace('"', '&quot;')

		return d

	@staticmethod
	def add_header(merged_xml, d):
		"""
		Add XML header to the intermediate merged Nmap XML report.

		:param merged_xml: intermediate merged Nmap XML report filename
		:type merged_xml: str
		:param d: dictionary containing XML metadata
		:type d: dict
		"""
		nmap_header = '<?xml version="1.0" encoding="UTF-8"?>'
		nmap_header += '<!DOCTYPE nmaprun>'
		nmap_header += '<?xml-stylesheet href="file:///usr/share/nmap/nmap.xsl" type="text/xsl"?>'
		nmap_header += '<!-- Nmap scan initiated by DivideAndScan: https://github.com/snovvcrash/DivideAndScan -->'
		nmap_header += '<!-- Nmap reports merged with nMapMerge.py: https://github.com/CBHue/nMap_Merger -->'
		nmap_header += f'''<nmaprun scanner="{d['scanner']}" args="(Example) {d['args']}" start="0" startstr="Thu Jan  1 00:00:00 1970" version="{d['version']}" xmloutputversion="{d['xmloutputversion']}">'''
		nmap_header += f'''<scaninfo type="{d['type']}" protocol="{d['protocol']}" numservices="0" services="0"/>'''
		nmap_header += '<verbose level="0"/>'
		nmap_header += '<debugging level="0"/>'

		with open(merged_xml, 'w') as fd:
			fd.write(nmap_header)

	@staticmethod
	def add_footer(merged_xml, n):
		"""
		Add XML footer to the merged Nmap XML report.

		:param merged_xml: merged Nmap XML report filename
		:type merged_xml: str
		:param n: number of hosts in merged Nmap XML report
		:type n: int
		"""
		nmap_footer = f'<runstats><finished time="0" timestr="Thu Jan  1 00:00:00 1970" elapsed="0" summary="Nmap done at Thu Jan  1 00:00:00 1970; {str(n)} IP address scanned in 0.0 seconds" exit="success"/>'
		nmap_footer += '</runstats>'
		nmap_footer += '</nmaprun>'

		with open(merged_xml, 'a') as mfd:
			mfd.write(nmap_footer)

	@staticmethod
	def merge_nmap(xml_file, merged_xml):
		"""
		Add another Nmap XML report contents to the intermediate merged XML report.

		:param xml_file: current Nmap XML report filename to be merged
		:type xml_file: pathlib.PosixPath
		:param merged_xml: intermediate merged Nmap XML report filename
		:type merged_xml: str
		:return: the number of hosts in current Nmap XML report
		:rtype: int
		"""
		with open(merged_xml, mode='a', encoding='utf-8') as mfd:
			with open(xml_file) as fd:
				nmap_xml, n = ET.parse(fd), 0
				for host in nmap_xml.findall('host'):
					chost = ET.tostring(host, encoding='unicode', method='xml')
					mfd.write(chost)
					mfd.flush()
					n += 1

		return n



#!/usr/bin/env python3

import sys
from pathlib import Path
from collections import namedtuple
from argparse import ArgumentParser, RawTextHelpFormatter, RawDescriptionHelpFormatter

from das.modules.add import AddMasscanOutput, AddRustscanOutput, AddNaabuOutput, AddNmapOutput
from das.modules.scan import ScanShow, ScanRun
from das.modules.report import NmapMerger
from das.modules.common import BANNER, Logger


def parse_args():
	"""
	Process arguments.

	:return: arguments namespace
	:rtype: argparse.ArgumentParser.Namespace
	"""
	parser = ArgumentParser(description=BANNER, formatter_class=RawTextHelpFormatter, epilog='Psst, hey buddy... Wanna do some organized p0r7 5c4nn1n6?')

	subparser = parser.add_subparsers(dest='subparser')

	add_epilog = """
	examples:

	  das add masscan '-e eth0 --rate 1000 -iL hosts.txt -p1-65535 --open'
	  das add rustscan '-b 1000 -t 2000 -u 5000 -a hosts.txt -r 1-65535 -g --no-config --scan-order "Random"'
	  das add -db testdb -rm naabu '-interface eth0 -rate 1000 -iL hosts.txt -p - -silent -s s'
	  das add nmap '-v -n -Pn --min-rate 1000 -T4 -iL hosts.txt -p1-49151 --open'
	""".replace('\t', '')
	add_parser = subparser.add_parser('add', formatter_class=RawDescriptionHelpFormatter, epilog=add_epilog, help='run a full port scan {masscan,rustscan,naabu,nmap} and add the output to DB')
	add_parser.add_argument('scanner_name', action='store', type=str, help='port scanner name')
	add_parser.add_argument('scanner_args', action='store', type=str, help='port scanner switches and options')
	add_parser.add_argument('-db', action='store', type=str, default='das', help='DB name to save the output into')
	add_parser.add_argument('-rm', action='store_true', default=False, help='drop the DB before updating its values')

	scan_epilog = """
	examples:

	  das scan -hosts all -show
	  das scan -hosts 192.168.1.0/24,10.10.13.37 -oA report1 -nmap '-Pn -sVC -O'
	  das scan -db testdb -ports 22,80,443,445 -oA report2 -parallel
	  das scan -db testdb -ports ports.txt -oA report2 -parallel -proc 4
	""".replace('\t', '')
	scan_parser = subparser.add_parser('scan', formatter_class=RawDescriptionHelpFormatter, epilog=scan_epilog, help='run targeted Nmap scans against hosts and ports from DB')
	scan_parser.add_argument('-db', action='store', type=str, default='das', help='DB name to retrieve the input from')
	scan_parser.add_argument('-nmap', action='store', type=str, default=None, help='custom Nmap options, so the final command will be "sudo nmap <OPTIONS> -oA scan/$output $ip -p$ports" (default is "sudo nmap -Pn -sV --version-intensity 6 -O -oA scan/$output $ip -p$ports")')
	group_parallel = scan_parser.add_argument_group('parallelism')
	group_parallel.add_argument('-parallel', action='store_true', default=False, help='run Nmap in multiple processes, number of processes is set with -p (-processes) argument')
	group_parallel.add_argument('-proc', action='store', type=int, default=None, help='number of parallel Nmap processes (if no value is provided, it will default to the number of processors on the machine)')
	group_action = scan_parser.add_mutually_exclusive_group(required=True)
	group_action.add_argument('-show', action='store_true', default=False, help='only show DB data, do not launch Nmap')
	group_action.add_argument('-oA', action='store', type=str, default=None, help='final report filename without extension (all formats: HTML, XML, simple text, grepable)')
	group_action.add_argument('-oX', action='store', type=str, default=None, help='final report filename without extension (XML+HTML formats)')
	group_action.add_argument('-oN', action='store', type=str, default=None, help='final report filename without extension (simple text format)')
	group_action.add_argument('-oG', action='store', type=str, default=None, help='final report filename without extension (grepable format)')
	group_criteria = scan_parser.add_mutually_exclusive_group(required=True)
	group_criteria.add_argument('-hosts', action='store', type=str, default=None, help='hosts to scan all their ports which were considered as open (a comma-separated string of IPs and/or CIDRs or a filename; "all" for all hosts in DB)')
	group_criteria.add_argument('-ports', action='store', type=str, default=None, help='ports to scan on every host where it was considered as open (a comma-separated string of ports or a filename; "all" for all ports in DB)')

	report_epilog = """
	examples:

	  das report -hosts all -show
	  das report -hosts 192.168.1.0/24,10.10.13.37 -oA report1
	  das report -ports 22,80,443,445 -oA report2
	  das report -ports ports.txt -oA report2
	""".replace('\t', '')
	report_parser = subparser.add_parser('report', formatter_class=RawDescriptionHelpFormatter, epilog=report_epilog, help='merge separate Nmap outputs into a single report in different formats')
	group_action = report_parser.add_mutually_exclusive_group(required=True)
	group_action.add_argument('-show', action='store_true', default=False, help='only show Nmap raw reports, do not merge into a file')
	group_action.add_argument('-oA', action='store', type=str, default=None, help='final report filename without extension (all formats: HTML, XML, simple text, grepable)')
	group_action.add_argument('-oX', action='store', type=str, default=None, help='final report filename without extension (XML+HTML formats)')
	group_action.add_argument('-oN', action='store', type=str, default=None, help='final report filename without extension (simple text format)')
	group_action.add_argument('-oG', action='store', type=str, default=None, help='final report filename without extension (grepable format)')
	group_criteria = report_parser.add_mutually_exclusive_group(required=True)
	group_criteria.add_argument('-hosts', action='store', type=str, help='hosts to add to report by IP (a comma-separated string of IPs and/or CIDRs or a filename; "all" for all host reports in Nmap directory)')
	group_criteria.add_argument('-ports', action='store', type=str, help='hosts to add to report by port (a comma-separated string of ports or a filename; "all" for all port reports in Nmap directory)')

	return parser.parse_args()


def main():
	"""
	Main function.
	"""
	args = parse_args()

	if len(sys.argv) == 1:
		print('usage: __main__.py [-h] {add,scan,report} ...\n')
		print(BANNER)
		sys.exit(0)

	logger = Logger()

	if args.subparser == 'add' or args.subparser == 'scan' and not args.show:
		logger.start_timer()

	if args.subparser == 'add':
		(Path.cwd() / '.db' / 'raw').mkdir(parents=True, exist_ok=True)

		if args.scanner_name == 'masscan':
			AddPortscanOutput = AddMasscanOutput
		elif args.scanner_name == 'rustscan':
			AddPortscanOutput = AddRustscanOutput
		elif args.scanner_name == 'naabu':
			AddPortscanOutput = AddNaabuOutput
		elif args.scanner_name == 'nmap':
			AddPortscanOutput = AddNmapOutput
		else:
			logger.print_error(f'{args.scanner_name}: Unsupported port scanner')
			sys.exit(1)

		P = Path.cwd() / '.db' / f'{args.db}.json'

		apo = AddPortscanOutput(str(P), args.rm, args.scanner_name, args.scanner_args)
		portscan_out, num_of_hosts = apo.parse()

		if P.exists():
			logger.print_info(f'Using DB -> {P.resolve()}')

		P = Path.cwd() / portscan_out
		if P.exists():
			logger.print_info(f'Raw port scanner output -> {P.resolve()}')

		logger.print_success(f'Successfully updated DB with {num_of_hosts} hosts')

	elif args.subparser == 'scan':
		(Path.cwd() / '.nmap').mkdir(exist_ok=True)

		output = {'oA': args.oA, 'oX': args.oX, 'oN': args.oN, 'oG': args.oG}

		P = Path.cwd() / '.db' / f'{args.db}.json'
		if P.exists():
			logger.print_info(f'Using DB -> {P.resolve()}')

		if args.show:
			ss = ScanShow(str(P), args.hosts, args.ports)
			if args.hosts:
				ss.nmap_by_hosts()
			elif args.ports:
				ss.nmap_by_ports()

		elif any(o for o in output.values()):
			Parallelism = namedtuple('Parallelism', 'enabled processes')
			parallel = Parallelism(args.parallel, args.proc)

			sr = ScanRun(str(P), args.hosts, args.ports)
			if args.hosts:
				sr.nmap_by_hosts(args.nmap, parallel)
			elif args.ports:
				sr.nmap_by_ports(args.nmap, parallel)

			nm = NmapMerger(args.hosts, args.ports, output)
			nm.generate()

	elif args.subparser == 'report':
		output = {'oA': args.oA, 'oX': args.oX, 'oN': args.oN, 'oG': args.oG}

		if args.show:
			nm = NmapMerger(args.hosts, args.ports)
			nm.show()

		elif any(o for o in output.values()):
			nm = NmapMerger(args.hosts, args.ports, output)
			nm.generate()

	if args.subparser == 'add' or args.subparser == 'scan' and not args.show:
		logger.stop_timer()


if __name__ == '__main__':
	main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['das', 'das.modules']

package_data = \
{'': ['*']}

install_requires = \
['defusedxml>=0.7.1,<0.8.0', 'netaddr>=0.8.0,<0.9.0', 'tinydb>=4.4.0,<5.0.0']

entry_points = \
{'console_scripts': ['das = das.divideandscan:main',
                     'divideandscan = das.divideandscan:main']}

setup_kwargs = {
    'name': 'divideandscan',
    'version': '0.1.8',
    'description': 'Divide full port scan results and use it for targeted Nmap runs',
    'long_description': '<p align="center">\n  <img src="https://user-images.githubusercontent.com/23141800/113610876-632a4300-9656-11eb-9583-d07f4e34d774.png" width="350px" alt="DivideAndScan">\n</p>\n\n<p align="center">\n  <strong>Divide <strike>Et Impera</strike> And Scan (and also merge the scan results)</strong>\n</p>\n\n<p align="center">\n  <a href="https://github.com/snovvcrash/DivideAndScan/blob/main/pyproject.toml#L3"><img src="https://img.shields.io/badge/version-0.1.8-success" alt="version" /></a>\n  <a href="https://github.com/snovvcrash/DivideAndScan/search?l=python"><img src="https://img.shields.io/badge/python-3.7-blue?logo=python&logoColor=white" alt="python" /></a>\n  <a href="https://www.codacy.com/gh/snovvcrash/DivideAndScan/dashboard?utm_source=github.com&amp;utm_medium=referral&amp;utm_content=snovvcrash/DivideAndScan&amp;utm_campaign=Badge_Grade"><img src="https://app.codacy.com/project/badge/Grade/35f0bdfece9846d7aab3888b01642813" alt="codacy" /></a>\n  <a href="https://github.com/snovvcrash/DivideAndScan/actions/workflows/publish-to-pypi.yml"><img src="https://github.com/snovvcrash/DivideAndScan/actions/workflows/publish-to-pypi.yml/badge.svg" alt="pypi" /></a>\n  <a href="https://github.com/snovvcrash/DivideAndScan/actions/workflows/publish-to-docker-hub.yml"><img src="https://github.com/snovvcrash/DivideAndScan/actions/workflows/publish-to-docker-hub.yml/badge.svg" alt="docker" /></a>\n</p>\n\n---\n\n**D**ivide**A**nd**S**can is used to efficiently automate port scanning routine by splitting it into 3 phases:\n\n1. Discover open ports for a bunch of targets.\n2. Run Nmap individually for each target with version grabbing and NSE actions.\n3. Merge the results into a single Nmap report (different formats available).\n\nFor the 1st phase a *fast* port scanner is intended to be used ([Masscan](https://github.com/robertdavidgraham/masscan) / [RustScan](https://github.com/RustScan/RustScan) / [Naabu](https://github.com/projectdiscovery/naabu)), whose output is parsed and stored in a single file database ([TinyDB](https://github.com/msiemens/tinydb)). Next, during the 2nd phase individual Nmap scans are launched for each target with its set of open ports (multiprocessing is supported) according to the database data. Finally, in the 3rd phase separate Nmap outputs are merged into a single report in different formats (XML / HTML / simple text / grepable) with [nMap_Merger](https://github.com/CBHue/nMap_Merger).\n\nPotential use cases:\n\n* Pentest engagements / red teaming with a large scope to enumerate.\n* Cybersecurity wargames / training CTF labs.\n* OSCP certification exam.\n\n## How It Works\n\n![how-it-works.png](https://user-images.githubusercontent.com/23141800/113610892-67566080-9656-11eb-8520-8fa2dcaf3463.png)\n\n## How to Install\n\n### Prerequisites\n\nTo successfully *divide and scan* we need to get some good port scanning tools.\n\n📑 **Note:** if you don\'t feel like messing with dependecies on your host OS, skip to the [Docker](#using-from-docker) part.\n\n#### Nmap\n\n```bash\nsudo apt install nmap sudo xsltproc -y\nsudo nmap --script-updatedb\n```\n\n#### Masscan\n\n```bash\ncd /tmp\ngit clone https://github.com/robertdavidgraham/masscan.git\ncd masscan\nmake\nsudo make install\ncd && rm -rf /tmp/masscan\n```\n\n#### RustScan\n\n```bash\ncd /tmp\n\nwget -qO- https://api.github.com/repos/RustScan/RustScan/releases/latest \\\n| grep "browser_download_url.*amd64.deb" \\\n| cut -d: -f2,3 \\\n| tr -d \\" \\\n| wget -qO rustscan.deb -i-\n\nsudo dpkg -i rustscan.deb\ncd && rm /tmp/rustscan.deb\n\nsudo wget https://gist.github.com/snovvcrash/8b85b900bd928493cd1ae33b2df318d8/raw/fe8628396616c4bf7a3e25f2c9d1acc2f36af0c0/rustscan-ports-top1000.toml -O /root/.rustscan.toml\n```\n\n#### Naabu\n\n```bash\nsudo mkdir /opt/projectdiscovery\ncd /opt/projectdiscovery\n\nwget -qO- https://api.github.com/repos/projectdiscovery/naabu/releases/latest \\\n| grep "browser_download_url.*linux-amd64.tar.gz" \\\n| cut -d: -f2,3 \\\n| tr -d \\" \\\n| sudo wget -qO naabu.tar.gz -i-\n\nsudo tar -xvzf naabu.tar.gz\nsudo mv naabu-linux-amd64 naabu\nsudo rm naabu.tar.gz README.md LICENSE.md\nsudo ln -vs /opt/projectdiscovery/naabu /usr/local/bin/naabu\n```\n\n### Installation\n\nDivideAndScan is available on PyPI as `divideandscan`, though I recommend installing it from GitHub with [pipx](https://github.com/pipxproject/pipx) in order to always have the bleeding-edge version:\n\n```console\n~$ pipx install -f "git+https://github.com/snovvcrash/DivideAndScan.git"\n~$ das\n```\n\nFor debbugging purposes you can set up a dev environment with [poetry](https://github.com/python-poetry/poetry):\n\n```console\n~$ git clone https://github.com/snovvcrash/DivideAndScan\n~$ cd DivideAndScan\n~$ poetry install\n~$ poetry run das\n```\n\n📑 **Note:** DivideAndScan uses sudo to run all the port scanners, so it will ask for the password when scanning commands are invoked.\n\n### Using from Docker\n\nYou can run DivideAndScan in a Docker container as follows:\n\n```console\n~$ docker run -it --rm --name das -v `pwd`:/app snovvcrash/divideandscan\n```\n\nSince the tool requires some input data and produces some output data, you should specify your current working directory as the mount point at `/app` within the container. You may want to set an alias to make the base command shorter:\n\n```console\n~$ alias das=\'docker run -it --rm --name das -v `pwd`:/app snovvcrash/divideandscan\'\n~$ das\n```\n\n## How to Use\n\n![how-to-use.png](https://user-images.githubusercontent.com/23141800/113610915-6fae9b80-9656-11eb-8b1a-db503dd43861.png)\n\n### 0. Preparations\n\n<table>\n<tr>\n<td>\n\nMake a new directory to start DivideAndScan from. The tool will create subdirectories in CWD to store the output, so I recommend launching it from a clean directory to stay organized:\n\n```console\n~$ mkdir divideandscan\n~$ cd divideandscan\n```\n\n</td>\n</tr>\n</table>\n\n### 1. Filling the DB\n\n<table>\n<tr>\n<td>\n\nProvide the `add` module a command for a fast port scanner to discover open ports in a desired range.\n\n⚠️ **Warning:** please, make sure that you understand what you\'re doing, because nearly all port scanning tools [can damage the system being tested](https://github.com/RustScan/RustScan/wiki/Usage#%EF%B8%8F-warning) if used improperly.\n\n```console\n# Masscan\n~$ das add masscan \'--rate 1000 -iL hosts.txt -p1-65535 --open\'\n# RustScan\n~$ das add rustscan \'-b 1000 -t 2000 -u 5000 -a hosts.txt -r 1-65535 -g --no-config\'\n# Naabu\n~$ das add naabu \'-rate 1000 -iL hosts.txt -p - -silent -s s\'\n# Nmap, -v flag is always required for correct parsing!\n~$ das add nmap \'-v -n -Pn --min-rate 1000 -T4 -iL hosts.txt -p1-65535 --open\'\n```\n\nWhen the module completes its work, a hidden directory `.db` is created in CWD containig the database file and raw scan results.\n\n</td>\n</tr>\n</table>\n\n### 2. Targeted Scanning\n\n<table>\n<tr>\n<td>\n\nLaunch targeted Nmap scans with the `scan` module. You can adjust the scan surface with either `-hosts` or `-ports` option:\n\n```console\n# Scan by hosts\n~$ das scan -hosts all -oA report1\n~$ das scan -hosts 192.168.1.0/24,10.10.13.37 -oA report1\n~$ das scan -hosts hosts.txt -oA report1\n# Scan by ports\n~$ das scan -ports all -oA report2\n~$ das scan -ports 22,80,443,445 -oA report2\n~$ das scan -ports ports.txt -oA report2\n```\n\nTo start Nmap simultaneously in multiple processes, specify the `-parallel` switch and set number of workers with the `-proc` option (if no value is provided, it will default to the number of processors on the machine):\n\n```console\n~$ das scan -hosts all -oA report -parallel [-proc 4]\n```\n\nThe output format is selected with `-oX`, `-oN`, `-oG` and `-oA` options for XML+HTML formats, simple text format, grepable format and all formats respectively. When the module completes its work, a hidden directory `.nmap` is created in CWD containig Nmap raw scan reports.\n\nAlso, you can inspect the contents of the database with `-show` option before actually launching the scans:\n\n```console\n~$ das scan -hosts all -show\n```\n\n</td>\n</tr>\n</table>\n\n### 3 (Optional). Merging the Reports\n\n<table>\n<tr>\n<td>\n\nIn order to generate a report independently of the `scan` module, you should use the `report` module. It will search for Nmap raw scan reports in the `.nmap` directory and process and merge them based on either `-hosts` or `-ports` option:\n\n```console\n# Merge outputs by hosts\n~$ das report -hosts all -oA report1\n~$ das report -hosts 192.168.1.0/24,10.10.13.37 -oA report1\n~$ das report -hosts hosts.txt -oA report1\n# Merge outputs by ports\n~$ das report -ports all -oA report2\n~$ das report -ports 22,80,443,445 -oA report2\n~$ das report -ports ports.txt -oA report2\n```\n\n📑 **Note:** keep in mind that the `report` module does **not** search the DB when processing the `-hosts` or `-ports` options, but looks for Nmap raw reports directly in `.nmap` directory instead; it means that `-hosts 127.0.0.1` argument value will be successfully resolved only if `.nmap/127-0-0-1.*` file exists, and `-ports 80` argument value will be successfully resolved only if `.nmap/port80.*` file exists.\n\n</td>\n</tr>\n</table>\n\n<details>\n<summary><strong>🔥 Example 🔥</strong></summary>\n\nLet\'s enumerate open ports for all live machines on [Hack The Box](https://www.hackthebox.eu/home/machines).\n\n1. Add mappings "host ⇄ open ports" to the database with Masscan. For demonstration purposes I will exclude dynamic port range to avoid unnecessary stuff by using `-p1-49151`. On the second screenshot I\'m reviewing scan results by hosts and by ports:\n\n```console\n~$ das add -db htb -rm masscan \'-e tun0 --rate 1000 -iL hosts.txt -p1-49151 --open\'\n```\n\n<p align="center">\n  <img src="https://user-images.githubusercontent.com/23141800/117919590-f578d300-b2f5-11eb-8afb-f8e3ed851e62.png" alt="example-1.png">\n</p>\n\n```console\n~$ das scan -db htb -hosts all -show\n~$ das scan -db htb -ports all -show\n```\n\n<p align="center">\n  <img src="https://user-images.githubusercontent.com/23141800/117919602-fa3d8700-b2f5-11eb-8d4a-f2edb0272e2e.png" alt="example-2.png">\n</p>\n\n2. Launch Nmap processes for each target to enumerate only ports that we\'re interested in (the open ports). On the second screenshot I\'m doing the same but starting Nmap processes simultaneously:\n\n```console\n~$ das scan -db htb -hosts all -oA report\n```\n\n<p align="center">\n  <img src="https://user-images.githubusercontent.com/23141800/117919624-03c6ef00-b2f6-11eb-9539-64a5a6ced1cf.png" alt="example-3.png">\n</p>\n\n```console\n~$ das scan -db htb -hosts all -oA report -nmap \'-Pn -sVC -O\' -parallel\n```\n\n<p align="center">\n  <img src="https://user-images.githubusercontent.com/23141800/117919633-0a556680-b2f6-11eb-8cbe-78d1e9ce16f1.png" alt="example-4.png">\n</p>\n\n3. As a result we now have a single report in all familiar Nmap formats (simple text, grepable, XML) as well as a pretty HTML report.\n\n<p align="center">\n  <img src="https://user-images.githubusercontent.com/23141800/117919635-0c1f2a00-b2f6-11eb-933f-ee812e6f6bd0.png" alt="example-5.png">\n</p>\n\n</details>\n\n## Help\n\n```\nusage: das [-h] {add,scan,report} ...\n\n -----------------------------------------------------------------------------------------------\n|  ________  .__      .__    .___        _____              .____________                       |\n|  \\______ \\ |__|__  _|__| __| _/____   /  _  \\   ____    __| _/   _____/ ____ _____    ____    |\n|   |    |  \\|  \\  \\/ /  |/ __ |/ __ \\ /  /_\\  \\ /    \\  / __ |\\_____  \\_/ ___\\\\__  \\  /    \\   |\n|   |    `   \\  |\\   /|  / /_/ \\  ___//    |    \\   |  \\/ /_/ |/        \\  \\___ / __ \\|   |  \\  |\n|  /_______  /__| \\_/ |__\\____ |\\___  >____|__  /___|  /\\____ /_______  /\\___  >____  /___|  /  |\n|          \\/                 \\/    \\/        \\/     \\/      \\/       \\/     \\/     \\/     \\/   |\n|  {@snovvcrash}            {https://github.com/snovvcrash/DivideAndScan}             {vX.Y.Z}  |\n -----------------------------------------------------------------------------------------------\n\npositional arguments:\n  {add,scan,report}\n    add              run a full port scan {masscan,rustscan,naabu,nmap} and add the output to DB\n    scan             run targeted Nmap scans against hosts and ports from DB\n    report           merge separate Nmap outputs into a single report in different formats\n\noptional arguments:\n  -h, --help         show this help message and exit\n\nPsst, hey buddy... Wanna do some organized p0r7 5c4nn1n6?\n```\n\n## ToDo\n\n* [x] Add [projectdiscovery/naabu](https://github.com/projectdiscovery/naabu) parser\n* [ ] Add armada scanner (?) parser\n\n## Support\n\nIf this tool has been useful for you, feel free to buy me a <strike>beer</strike> coffee!\n\n[![beer.png](https://user-images.githubusercontent.com/23141800/113611163-d03dd880-9656-11eb-9279-e5e2689b0c1b.png)](https://buymeacoff.ee/snovvcrash)\n',
    'author': 'Sam Freeside',
    'author_email': 'snovvcrash@protonmail.ch',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/snovvcrash/DivideAndScan',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python3

import os
import socket
from pathlib import Path
from abc import abstractmethod
from collections import defaultdict
from tempfile import NamedTemporaryFile
from multiprocessing import current_process
from concurrent.futures import ProcessPoolExecutor

from tinydb import TinyDB, Query
from netaddr import IPNetwork

from das.modules.common import Logger


class ScanBase:
	"""Base class for searching DB and/or initiating Nmap scans."""

	def __init__(self, db, hosts, ports):
		"""
		Constructor.

		:param db: a tinydb database file path
		:type db: tinydb.TinyDB
		:param hosts: a list of hosts to interact with ("all" for all the hosts in DB)
		:type hosts: list
		:param ports: a list of ports to interact with ("all" for all the ports in DB)
		:type ports: list
		:return: base class object
		:rtype: das.modules.scan.ScanBase
		"""
		self.db = TinyDB(db)
		self.Host = Query()
		self.total_scans = 0

		if hosts:
			try:
				if Path(hosts).exists():
					with open(hosts, 'r') as fd:
						hosts = ','.join(i.strip() for i in fd.read().splitlines())
			except OSError:  # catching [Errno 36] File name too long
				pass

			if hosts == 'all':
				result = self.db.all()
			else:
				hosts = hosts.split(',')
				hosts = [IPNetwork(h) for h in hosts]
				hosts = [str(ip) for ip_obj in hosts for ip in ip_obj]
				result = self.db.search(self.Host.ip.one_of(hosts))

			self.ip_dict = defaultdict(set)
			for item in result:
				self.ip_dict[item['ip']].add(item['port'])

			self.total_scans += len(self.ip_dict)

		elif ports:
			try:
				if Path(ports).exists():
					with open(ports, 'r') as fd:
						ports = ','.join(i.strip() for i in fd.read().splitlines())
			except OSError:  # catching [Errno 36] File name too long
				pass

			if ports == 'all':
				result = self.db.all()
			else:
				ports = [int(p) for p in ports.split(',')]
				result = self.db.search(self.Host.port.one_of(ports))

			self.port_dict = defaultdict(set)
			for item in result:
				self.port_dict[item['port']].add(item['ip'])

			self.total_scans += len(self.port_dict)

		Logger.print_info(f'Total scans -> {self.total_scans}')

	@abstractmethod
	def nmap_by_hosts(self):
		"""Interface for a DB host searching method."""
		raise NotImplementedError

	@abstractmethod
	def nmap_by_ports(self):
		"""Interface for a DB port searching method."""
		raise NotImplementedError


class ScanShow(ScanBase):
	"""Child class for searching through DB and printing the results."""

	def nmap_by_hosts(self):
		"""Search DB by hosts and print mapping "live_host -> [open_ports]". No Nmap scan is launched."""
		for ip, ports in sorted(self.ip_dict.items(), key=lambda x: socket.inet_aton(x[0])):
			sorted_ports = ','.join([str(p) for p in sorted(ports)])
			Logger.print_success(f'IP {ip} ({len(ports)}) -> [{sorted_ports}]')

	def nmap_by_ports(self):
		"""Search DB by ports and print mapping "open_port -> [live_hosts]". No Nmap scan is launched."""
		for port, ip_list in sorted(self.port_dict.items()):
			sorted_ip_list = ','.join(sorted(ip_list, key=socket.inet_aton))
			Logger.print_success(f'Port {port} ({len(ip_list)}) -> [{sorted_ip_list}]')


class ScanRun(ScanBase):
	"""Child class for initiating Nmap scans."""

	def nmap_by_hosts(self, nmap_opts, parallel):
		"""
		Search DB by hosts and launch Nmap scans for mappings "live_host -> [open_ports]".

		:param nmap_opts: custom Nmap options that will replace the default ones
		:type nmap_opts: str
		:param parallel: namedtuple('Parallelism', 'enabled processes')
		:type parallel: collections.namedtuple
		"""
		nmap_commands, i = [], 1
		for ip, ports in sorted(self.ip_dict.items(), key=lambda x: socket.inet_aton(x[0])):
			if not parallel.enabled:
				Logger.print_separator(f'IP: {ip}', prefix=f'{i}/{self.total_scans}')

			nmap_out = ip.replace('.', '-')
			sorted_ports = ','.join([str(p) for p in sorted(ports)])

			if nmap_opts is None:
				cmd = f"""sudo nmap -Pn -sV --version-intensity 6 -O -oA .nmap/{nmap_out} {ip} -p{sorted_ports}"""
			else:
				cmd = f"""sudo nmap {nmap_opts} -oA .nmap/{nmap_out} {ip} -p{sorted_ports}"""

			if parallel.enabled:
				cmd += ' > /dev/null 2>&1'
				nmap_commands.append(cmd)
			else:
				Logger.print_cmd(cmd)
				os.system(cmd)

			i += 1

		if parallel.enabled:
			with ProcessPoolExecutor(max_workers=parallel.processes) as executor:
				executor.map(nmap, nmap_commands)

	def nmap_by_ports(self, nmap_opts, parallel):
		"""
		Search DB by ports and launch Nmap scans for mappings "open_port -> [live_hosts]".

		:param nmap_opts: custom Nmap options that will replace the default ones
		:type nmap_opts: str
		:param parallel: namedtuple('Parallelism', 'enabled processes')
		:type parallel: collections.namedtuple
		"""
		nmap_commands, i = [], 1
		for port, ip_list in sorted(self.port_dict.items()):
			if not parallel.enabled:
				Logger.print_separator(f'Port: {port}', prefix=f'{i}/{self.total_scans}')

			nmap_out = f'port{port}'
			with NamedTemporaryFile('w+') as tmp:
				tmp.writelines(f'{i}\n' for i in ip_list)
				tmp.seek(0)

				if nmap_opts is None:
					cmd = f"""sudo nmap -Pn -sV --version-intensity 6 -O -oA .nmap/{nmap_out} -iL {tmp.name} -p{port}"""
				else:
					cmd = f"""sudo nmap {nmap_opts} -oA .nmap/{nmap_out} -iL {tmp.name} -p{port}"""

				sorted_ip_list = ','.join(sorted(ip_list, key=socket.inet_aton))
				Logger.print_cmd(f'{tmp.name}: {sorted_ip_list}')

				if parallel.enabled:
					cmd += ' > /dev/null 2>&1'
					nmap_commands.append(cmd)
				else:
					Logger.print_cmd(cmd)
					os.system(cmd)

				i += 1

		if parallel.enabled:
			with ProcessPoolExecutor(max_workers=parallel.processes) as executor:
				executor.map(nmap, nmap_commands)


def nmap(command):
	"""
	Helper function to run multiple Nmap processes in parallel.

	:param command: Nmap command to execute
	:type command: str
	"""
	Logger.print_cmd(command, parallel=current_process().name)
	os.system(command)
